/*****************************************************************************

   (c) 1996 by Salvador Eduardo Tropea (SET).
   e-mail: salvador@inti.edu.ar

   You can use this code for any purpose if you don't remove the original
copyright.

   Filename -       Inf.cpp

   Functions:
                    Member functions of following classes
                        TInfViewer
                        TInfWindow
                        THelpControlDialog
                        TStrListBox

History: (for Inf.Cpp and InfBase.Cpp)
5/5/96 v0.0.7
Added: BestMatch.
       Robert modifications for cursor visibility and use.
       Local help in ^H.
       Support for a StatusLine.
Modifyed: ALT-F10 and ALT-F1 now are commands.
          Double line links with speed optimization.
7/5/96
Modified: getLine of TInfTopic from char *getLine(int) to void getLine(int,
          char *) for 2 reasons: 1) It's more safe 2) supress an extra
          strcpy in TInfViewer::draw().

12/6/96
Modified: selWord=0 in draw() to avoid warnings. (Robert)
Added:    drawView(); switchToTopic because some times the window isn't
          updated (when there is no cursor movement between topics). (Robert)
Modified: TInfIndicator was replaced by a change of title. (Robert)
Modified: An error in MakeVisible (-1).

??/10/96
Added:    Support for compressed files.
Added:    International support.

04/11/96 v0.0.8
Modified: Parser of menus and * aaaa: aaa.XXXXXX.

12/11/96 v0.0.9
Modified: to hide the *Note string.

13/11/96
Modified: the methode to hide *Notes.
Added: A little selection and paste to the clipboard feature.

1/12/96
Added: isTheOne to manage the info from the standalone editor.
Modified: Now getTopic is safer when the file doesn't exist.

9/12/96
Added: Match of topic while writing the initials. (requested by NIK).

10/12/96
Bug Killed: '0' didn't check if the number of refs. was 0. (reported by Artur).
Added: A help context to each dialog. (requested by Artur).

4/1/97
Added: Now the selection disapears when we change of topic.
Added: Selection in reverse direction.
Added: New behaviour of End.
Added: Selection with End,Home,Page Up and Page Down.
Added: Mouse selection.

9/1/97
Added: Support for files without Tag Table (just plain text files with delimiters).

10/1/97
Added: Better support for Tabs.
Added: Ported to Linux (no shift arrows, double click emulated).
Fixed: Now the items in the first line below the Menu entry is taked in count.
Added: A dialog to choose any node of the file.
v0.1.0

ToDo:
  When pressing the initials accumulate the keystrokes, works but needs some
adjustments, for example I didn't take care about names crossing lines.
  Enhance the selection.

*****************************************************************************/

#if defined(__GNUC__) && !defined(__DJGPP__)
extern "C" char *strupr(char *);
#define Linux
#define NO_DOUBLE_CLICK
#endif

#ifdef FOR_EDITOR
// That's the first include because is used to configure the editor.
#include "ceditint.h"
#endif

#define Uses_TStreamableClass
#define Uses_TPoint
#define Uses_TStreamable
#define Uses_TRect
#define Uses_TScrollBar
#define Uses_TScroller
#define Uses_TDrawBuffer
#define Uses_TEvent
#define Uses_TWindow
#define Uses_TKeys
#define Uses_TPalette
#define Uses_TButton
#define Uses_TListBox
#define Uses_TLabel
#define Uses_TDialog
#define Uses_MsgBox
#define Uses_TFindDialogRec
#define Uses_TInputLine
#define Uses_TCheckBoxes
#define Uses_TRadioButtons
#define Uses_THistory
#define Uses_TSItem
#define Uses_TStatusLine
#define Uses_TFrame
#define Uses_TStringCollection
#define Uses_TSortedListBox
#include <tv.h>

#include "tnocastc.h"

#if !defined( __Inf_H )
#include "inf.h"
#endif  // __Inf_H

#if !defined( __UTIL_H )
#include "util.h"
#endif  // __UTIL_H

#if !defined( __STRING_H )
#include <string.h>
#endif  // __STRING_H

#if !defined( __LIMITS_H )
#include <limits.h>
#endif  // __LIMITS_H

#if !defined( __STAT_H )
#include <sys/stat.h>
#endif  // __STAT_H

#if !defined( __CTYPE_H )
#include <ctype.h>
#endif  // __CTYPE_H

#ifndef Linux
# if !defined( __IO_H )
#  include <io.h>
# endif  // __IO_H
#else
# include <stdlib.h>
#endif

#pragma warn -dsz

#define DEBUG
#ifdef DEBUG
# ifndef FOR_EDITOR
# include "nointl.h"
# endif
#endif

extern execDialog( TDialog *d, void *data );
const char * const TCollectionNoSort::name="TCollectionNoSort";
TDialog *createNodeChooseDialog();

#if defined(__GNUC__) && defined(__DJGPP__)
const
  smExtend = 0x01,
  smDouble = 0x02;
#endif
const
  smEndSel = 0x04,
  smStartSel=0x08;

#ifndef __GNUC__
static unsigned short getshiftstate(void)
{
 const uchar far *const shiftState = (uchar far *)MK_FP( 0x40, 0x17 );
 return *shiftState;
}
#else
# ifdef __DJGPP__
#  define getshiftstate()  event.keyDown.shiftState
# else
#  define getshiftstate()  0
# endif
#endif


#define kbFlagsShift (kbRightShift | kbLeftShift)
#define kbCtrlFlg    kbCtrlShift
#define kbAltFlg     kbAltShift
#define kbShiftFlg   kbFlagsShift

/***************************************************************
 class TInfViewer
***************************************************************/

TInfViewer::TInfViewer( const TRect& bounds, TScrollBar* aHScrollBar,
    TScrollBar* aVScrollBar, TInfFile *aInfFile, char *context,
    char *match )
    : TScroller( bounds, aHScrollBar, aVScrollBar )
{
 char NameCopy[MAX_NODE_NAME];

 // pumm!!! bug killed, fucking strings in code by GCC, i need a buffer not
 // a simple string (my fault).
 strcpy(NameCopy,context);

 lockCount=0;
 mustBeRedrawed=False;

 options |= ofSelectable;
 growMode = gfGrowHiX | gfGrowHiY;
 hFile = aInfFile;
 topic = aInfFile->getTopic(NameCopy);
 topic->setWidth(size.x);
 setLimit(topic->maxWidth() + size.x, topic->numLines() + size.y);
 QuickLen=0;

 int LaPego;
 if ((selected=topic->selBestMatch(match,LaPego))!=0)
//    selected = 1;
// else
   {
    cursor.x=topic->crossRefs[selected-1].offset;
    int y=topic->crossRefs[selected-1].line-1;
    if (y>(size.y>>1))
      {
       delta.y=y-(size.y>>1);
       cursor.y=size.y>>1;
      }
    else
      {
       delta.y=y;
       cursor.y=1;
      }
    delta.x=0;
    scrollTo(cursor.x,y);
   }

 TextSelected = False;
 findStr[0]=0;
 findType=0;
 findWhere=fitInTopic;
 SearchArmed=0;
 lastSearch=0;
 OpenVerbose=1;

 // History
 HistPoint=0;
 History= new TCollectionNoSort(10,10);
 stHisto *h = new stHisto;
 strcpy(h->Name,"(");
 strcat(h->Name,hFile->NameOfFile);
 strcat(h->Name,")");
 strcat(h->Name,context);
 h->HisPos=cursor+delta;
 h->HisSel=selected;
 History->insert((void *)h);

 state |= sfCursorVis;

 if (LaPego)
   {
    switchToTopic(topic->crossRefs[selected-1].Name);
    cursor.x=delta.x=cursor.y=delta.y=0;
   }

 SetTitle( hFile->NameOfFile, topic->Node );

 selRowStart=selRowEnd=selColStart=selColEnd=0;
 selecting=False;
 InsertRoutine=NULL;
}


TInfViewer::~TInfViewer()
{
 delete hFile;
 delete topic;
 delete History;
}

static int PointInCrossRef(TPoint point,TInfTopic *topic,int & i)
{
 int crossCount = topic->numRefs;
 i = 0;
 if (crossCount == 0) return -1;
 point.y++;
 for (i=0;i<crossCount;i++)
    {
     CrossRef & ref = topic->crossRefs[i];
     if (ref.line > point.y) return -1;
     // Support for crossRefs that cross lines, slow
     if (ref.line == (point.y-1))
       {
        if (ref.linebreak && point.x<=ref.linebreak)
           return i;
        continue;
       }
     if (ref.line < point.y) continue;
     if (ref.offset > point.x) return -1;
     if (ref.offset+ref.length <= point.x) continue;
     return i;
    }
 i = crossCount;
 return -1;
}

void TInfViewer::resetCursor()
{
  int before;
  TScroller::resetCursor();
  selected = PointInCrossRef(cursor+delta,topic,before)+1;
}

void TInfViewer::scrollDraw()
{
 TPoint  d;

 if( hScrollBar != 0 )
     d.x = hScrollBar->value;
 else
     d.x = 0;

 if( vScrollBar != 0 )
     d.y = vScrollBar->value;
 else
     d.y = 0;

 if ( (d.x - delta.x) != cursor.x || (d.y - delta.y) != cursor.y)
 {
   if (d.x - delta.x >= size.x)
   {
     delta.x = d.x - size.x + 1;
   }
   else if (d.x - delta.x < 0)
   {
     delta.x = d.x;
   }
   if (d.y - delta.y >= size.y)
   {
     delta.y = d.y - size.y + 1;
   }
   else if (d.y - delta.y < 0)
   {
     delta.y = d.y;
   }
   setCursor(d.x-delta.x,d.y-delta.y);
   if( drawLock != 0 )
       drawFlag = True;
   else
       drawView();
 }
}

void TInfViewer::changeBounds( const TRect& bounds )
{
 TScroller::changeBounds(bounds);
 topic->setWidth(size.x);
 setLimit(topic->maxWidth() + size.x, topic->numLines() + size.y);
}

static void moveChar(int indent, char c, char attr, int count, ushort *b)
{
 ushort val=c | (attr<<8);

 for (b+=indent; count; count--, b++)
     *b=val;
}

static void moveStr(int indent, char *buffer, char attr, ushort *b)
{
 char *s=(char *)(b+indent);

 while (*buffer)
   {
    *s=*buffer;
    *(s+1)=attr;
    s+=2;
    buffer++;
   }
}

void TInfViewer::draw()
{
 if (lockCount)
   {
    mustBeRedrawed=True;
    return;
   }
 mustBeRedrawed=False;

 ushort b[256];
 char linea[256],*line;
 char buffer[256];
 char *bufPtr;
 unsigned char aux;
 int i, j, l, k;
 int keyCount;
 ushort normal, keyword, selKeyword, c,selWord=0,colSelArea;
 TPoint keyPoint;
 uchar keyLength;
 char *keyRef;
 int NumCrossRefs=topic->getNumCrossRefs();
 int LineaReal;
 int Exceso;
 char *bf=(char *)b;

 normal = getColor(1);
 keyword = getColor(2);
 selKeyword = getColor(3);
 if (TextSelected)
    selWord = getColor(4);
 colSelArea=0x70;
 keyCount = 0;
 keyPoint.x = 0;
 keyPoint.y = 0;
 topic->setWidth(size.x);

 // Saltear las crossrefs que estn antes en la pantalla
 if (NumCrossRefs > 0)
     {
      do
       {
        topic->getCrossRef(keyCount++, keyPoint, keyLength, keyRef, Exceso);
       }
      while ( (keyCount <= NumCrossRefs) && (keyPoint.y <= delta.y));
     }

 for (i = 1, LineaReal=delta.y+1; i <= size.y; ++i,LineaReal++)
     {
     topic->getLine(LineaReal,linea);
     if (linea[0]==30)
       {
        aux=(unsigned char)linea[1];
        if (aux<='9')
           c=(aux-'0')<<4;
        else
           c=(aux-(unsigned char)'A'+10)<<4;
        aux=(unsigned char)linea[2];
        if (aux<='9')
           c+=aux-'0';
        else
           c+=aux-(unsigned char)'A'+10;
        line=linea+3;
       }
     else
       {
        c=normal;
        line=linea;
       }
     moveChar(0, ' ', c, size.x, b); // Attr pedido
       
     if ((int)strlen(line) > delta.x)
         {
         bufPtr = line + delta.x;
         strncpy(buffer, bufPtr, size.x);
         buffer[size.x] = 0;
         moveStr(0, buffer, c, b);
         }
     else
         moveStr(0, "", c, b);


     while (LineaReal == keyPoint.y)
         {
         if (Exceso>0)
            l = keyLength-Exceso;
         else
            l = keyLength;
         if (keyPoint.x < delta.x )
           {
            l -= (delta.x - keyPoint.x);
            keyPoint.x = delta.x;
           }
         if (keyCount == selected)
            c = selKeyword;
         else
            c = keyword;
         for (j = 0; j < l; ++j)
             bf[(keyPoint.x - delta.x + j)*2+1]=c;
         if (Exceso>0)
           {
            keyPoint.x=0;
            keyPoint.y++;
            keyLength=Exceso+1;
            Exceso=0;
           }
         else
           {
            ++keyCount;
            if (keyCount <= NumCrossRefs)
                topic->getCrossRef(keyCount-1, keyPoint, keyLength, keyRef, Exceso);
            else
                keyPoint.y = 0;
           }
         }

     // Paint the selection
     if (LineaReal>=selRowStart && LineaReal<=selRowEnd)
       {
        int start;
        if (selRowStart==selRowEnd)
          {
           l=selColEnd-selColStart;
           if ( selColStart<delta.x )
             {
              l -= (delta.x - selColEnd);
              start = 0;
             }
           else
              start = selColStart - delta.x;
          }
        else
          {
           if (LineaReal==selRowStart)
             {
              start=selColStart-delta.x;
              l=size.x-start;
             }
           else
             if (LineaReal==selRowEnd)
               {
                start=0;
                l=selColEnd-delta.x;
               }
             else
               {
                start=0;
                l=size.x;
               }
          }
        for(j = 0, k=start*2+1; j < l; ++j, k+=2)
            bf[k] = (bf[k] & 0x0F) | colSelArea;
       }

     // Paint the founded word
     if (TextSelected && (LineaReal>=LineStartSelect && LineaReal<=LineEndSelect))
       {
        int start;
        l = ColEndSelect-ColStartSelect;
        if ( ColStartSelect<delta.x )
          {
           l -= (delta.x - ColStartSelect);
           start = 0;
          }
        else
           start = ColStartSelect - delta.x;
        for(j = 0, k=start*2+1; j < l; ++j, k+=2)
            bf[k]=selWord;
       }
     writeLine(0, i-1, size.x, 1, b);
     }
}

void TInfViewer::SetTitle(char *File, char *Node)
{
 if (owner) /* is a window ? */
   {
    char *newTitle = new char[25+strlen(File)+strlen(Node)+2];

    if (newTitle)
      {
       sprintf(newTitle,_("InfView - File: %s - Node: %s"),File,Node);

       delete (char *)((TWindow *)owner)->title;
       (char *)((TWindow *)owner)->title = newTitle;
       ((TWindow *)owner)->frame->drawView();
      }
   }
}


TPalette& TInfViewer::getPalette() const
{
    static TPalette palette(cInfViewer, sizeof( cInfViewer)-1);
    return palette;
}

void TInfViewer::makeSelectVisible()
{
 int ProxLine;
 TPoint keyPoint;
 uchar keyLength;
 char *keyRef;

 topic->getCrossRef(selected-1, keyPoint, keyLength, keyRef, ProxLine);
 if (QuickLen)
    keyPoint.x+=QuickVisPos-1;
 makeVisible(keyPoint,keyLength);
}


void TInfViewer::makeVisible( TPoint& keyPoint, int largo )
{
 if ((keyPoint.x+largo)>(delta.x+size.x) && (size.x>=largo))
    scrollTo(keyPoint.x+largo-1,keyPoint.y-1);
 else
    scrollTo(keyPoint.x,keyPoint.y-1);
}

void TInfViewer::switchToTopic( char *keyRef, TPoint Pos )
{
 char NameCopy[MAX_NODE_NAME];

 // Kill the selection
 selRowStart=selRowEnd=selColStart=selColEnd=0;
 selecting=False;

 // Because i can destroy the original bellow, keyRef is from topic and i go
 // to detroy it.
 strcpy(NameCopy,keyRef);

 if (topic != 0)
     delete topic;
 topic = hFile->getTopic(NameCopy,OpenVerbose);

 SetTitle( hFile->NameOfFile, topic->Node );

 topic->setWidth(size.x);
 setLimit(topic->maxWidth() + size.x, topic->numLines() + size.y);
 scrollTo(Pos.x,Pos.y);
 drawView();
 // No search
 TextSelected=False;
 lastSearch=topic->Text;
}

void TInfViewer::switchToTopic( char *keyRef )
{
 TPoint Pos;

 AddToHistory(keyRef);
 Pos.x=0; Pos.y=0;
 selected = 0;
 switchToTopic( keyRef,Pos );
}

void TInfViewer::switchToTopic( stHisto *h )
{
 AddToHistory(h->Name);
 selected=h->HisSel;
 switchToTopic(h->Name,h->HisPos);
}

static void AddFileToNode(char *dest, char *source, char *file)
{
 if (source[0]=='(')
    strcpy(dest,source);
 else
   {
    dest[0]='(';dest[1]=0;
    strcat(dest,file);
    strcat(dest,")");
    strcat(dest,source);
   }
}

void TInfViewer::AddToHistory(char *context)
{
 stHisto *h = (stHisto *)History->at(HistPoint);

 h->HisPos = cursor + delta;
 h->HisSel=selected;

 h = new stHisto;

 AddFileToNode(h->Name,context,hFile->NameOfFile);

 History->insert(h);
 HistPoint++;

 if (HistPoint>=MAX_HIST_DEEP)
   {
    History->atRemove(0);
    HistPoint--;
   }
}

char *TInfViewer::TakeFromHistory(TPoint& Pos)
{
 stHisto *h;
 if (HistPoint>0)
   {
    History->atRemove(HistPoint--);
    h=(stHisto *)History->at(HistPoint);
   }
 else
    h=(stHisto *)History->at(0);

 selected=h->HisSel;
 Pos = h->HisPos;
 return h->Name;
}



void TInfViewer::PasteToClipboard(void)
{
 char b[256];
 int deltaY=selRowEnd-selRowStart+1;

 if (deltaY<1 || (deltaY==0 && selColEnd<=selColStart))
    return;

 char *d=new char[deltaY*256];
 char *s=d;

 if (d==NULL)
    return;

 int i;
 int l=0,lTotal=0,lLine;
 int Xs=selColStart,Xe=selColEnd;
 for (i=selRowStart; i<=selRowEnd; i++)
    {
     if (i==selRowStart || i==selRowEnd)
       {
        lLine=topic->getLine(i,b);
        if (selRowStart==selRowEnd)
          {
           if (Xs>lLine) return;
           if (Xe>lLine) Xe=lLine;
           lTotal=Xe-Xs;
           memcpy(s,b+Xs,lTotal);
          }
        else
          if (i==selRowStart)
            {
             if (Xs<lLine)
               {
                lTotal=lLine-Xs;
                memcpy(s,b+Xs,lTotal);
                s+=lTotal;
                s[0]='\r';
                s[1]='\n';
                s+=2;
                lTotal+=2;
               }
            }
          else
            if (i==selRowEnd)
              {
               if (Xe>lLine)
                  Xe=lLine;
               l=Xe;
               memcpy(s,b,l);
               lTotal+=l;
              }
       }
     else
       {
        lLine=topic->getLine(i,s);
        s+=lLine;
        s[0]='\r';
        s[1]='\n';
        s+=2;
        lTotal+=lLine+2;
       }
    }

 if (lTotal)
    InsertRoutine(d,lTotal);

 delete d;
}

void TInfViewer::GoEndOfLine(void)
{
 char linea[256];
 int Xp=0,x=0;

 topic->getLine(delta.y+cursor.y+1,linea);
 while (linea[x])
   {
    if (!isspace(linea[x])) Xp=x;
    x++;
   }
 Xp++;
 scrollTo(Xp,delta.y+cursor.y);
}

void TInfViewer::UpdateSelection(int X, int Y)
{
 if (Y>selRowStartPoint || (Y==selRowStartPoint && X>=selColStartPoint))
   {
    selRowEnd=Y;
    selColEnd=X;
    selRowStart=selRowStartPoint;
    selColStart=selColStartPoint;
   }
 else
   {
    selRowStart=Y;
    selColStart=X;
    selRowEnd=selRowStartPoint;
    selColEnd=selColStartPoint;
   }
}

void TInfViewer::MoveToMouse( TPoint m, uchar selMode )
{
#ifdef NO_DOUBLE_CLICK
 int oldSel=selected;
#endif
 TPoint mouse = makeLocal( m );

 scrollTo(mouse.x+delta.x,mouse.y+delta.y);
 if (selMode & smExtend)
   { // Extends the selection
    UpdateSelection(cursor.x+delta.x,cursor.y+delta.y+1);
   }
 else
   { // Starts a selection
    if (selMode & smDouble)
      { // Jump to the topic
       if (selected > 0)
          switchToTopic(topic->getCrossRef(selected-1));
      }
    else
      {
       selRowStartPoint=cursor.y+delta.y+1;
       selColStartPoint=cursor.x+delta.x;
#ifdef NO_DOUBLE_CLICK
       if ((oldSel==selected) && selected>0)
          switchToTopic(topic->getCrossRef(selected-1));
#endif
      }
   }
}

void TInfViewer::unlock()
{
 if ( lockCount > 0 )
   {
    lockCount--;
    if (lockCount==0 && mustBeRedrawed)
       drawView();
   }
}

void TInfViewer::handleEvent( TEvent& event )
{
 uchar selectMode = 0;
 int S;
 int key;
 int X,Y;

 TScroller::handleEvent(event);
 switch (event.what)
     {
     case evCommand:
          switch (event.message.command)
             {
              case cmInfHelp:
                   switchToTopic("(infview)");
                   break;

              case cmInfControl:
                   {
                    int Comando =
                           execDialog( new THelpControlDialog(History), NULL );
                    if (Comando>=chcdHistSel)
                      {
                       switchToTopic((stHisto *)History->at(Comando-chcdHistSel));
                      }
                    else
                       switch (Comando)
                         {
                          case chcdNext: switchToTopic(topic->Next);
                                         break;
                          case chcdPrev: switchToTopic(topic->Prev);
                                         break;
                          case chcdUp:   switchToTopic(topic->Up);
                                         break;
                          case chcdPrevH: {
                                           TPoint Pos;
                                           char *contexto=TakeFromHistory(Pos);
                                           switchToTopic(contexto,Pos);
                                          }
                                          break;
                         }
                   }
                   break;

              case cmInfBack:
                   {
                    TPoint Pos;
                    char *contexto=TakeFromHistory(Pos);
                    switchToTopic(contexto,Pos);
                   }
                   break;

              default:
                   return;
             }
          break;

     case evKeyDown:
         S=getshiftstate();

         if (selecting)
           {
            if ((S & kbFlagsShift) != 0)
               selectMode = smExtend;
            else
  	      {
     	       selectMode = smEndSel;
     	       selecting = False;
  	      }
           }
         else
           {
            if ((S & kbFlagsShift) != 0)
  	       selectMode = smStartSel;
           }

         key=event.keyDown.keyCode;
         X=delta.x+cursor.x;
         Y=delta.y+cursor.y;
         if (selectMode==smStartSel && (key==kbDown || key==kbUp || key==kbLeft ||
             key==kbRight || key==kbEnd || key==kbHome || key==kbPgUp || key==kbPgDn))
           {
            selRowStartPoint=Y+1;
            selColStartPoint=X;
            selecting=True;
           }

         switch (key)
             {
             case kbTab:
                  if (PointInCrossRef(cursor+delta,topic,selected) >= 0)
                     selected++;
                  selected++;
                  if (selected > topic->getNumCrossRefs())
                      selected = 1;
                  if ( topic->getNumCrossRefs() != 0 )
                      makeSelectVisible();
                  QuickLen=0;
                  break;

             case kbShiftTab:
                  PointInCrossRef(cursor+delta,topic,selected);
                  if (selected == 0)
                      selected = topic->getNumCrossRefs();
                  if ( topic->getNumCrossRefs() != 0 )
                      makeSelectVisible();
                  QuickLen=0;
                  break;

             case kbEnter:
                  if (selected > 0)
                  if (selected <= topic->getNumCrossRefs())
                     switchToTopic(topic->getCrossRef(selected-1));
                  QuickLen=0;
                  break;

             case kbEsc:
                  if (owner->state & sfModal)
                     endModal(cmCancel);
                  else
                    {
                     event.what = evCommand;
                     event.message.command = cmClose;
                     putEvent(event);
                    }
                 QuickLen=0;
                 break;

             case kbDown:
                  if (selecting)
                    {
                     UpdateSelection(X,Y+2);
                     drawView();
                    }
                  QuickLen=0;
                  return;

             case kbUp:
                  if (selectMode)
                    {
                     UpdateSelection(X,Y);
                     drawView();
                    }
                  QuickLen=0;
                  return;

             case kbLeft:
                  if (selectMode)
                    {
                     UpdateSelection(X-1,Y+1);
                     drawView();
                    }
                  QuickLen=0;
                  return;

             case kbRight:
                  if (selectMode)
                    {
                     UpdateSelection(X+1,Y+1);
                     drawView();
                    }
                  QuickLen=0;
                  return;

             case kbCtrlIns:
                  if (InsertRoutine)
                     PasteToClipboard();
                  break;

             case kbEnd:
                  GoEndOfLine();
                  X=delta.x+cursor.x;
                  if (selectMode)
                     UpdateSelection(X,Y+1);
                  hScrollBar->setValue(X);
                  QuickLen=0;
                  break;

             case kbHome:
                  if (selectMode)
                    {
                     UpdateSelection(0,Y+1);
                     drawView();
                    }
                  QuickLen=0;
                  return;

             case kbPgUp:
                  scrollTo(X,Y-(size.y-1));
                  Y=cursor.y+delta.y;
                  if (selectMode)
                     UpdateSelection(cursor.x+delta.x+1,Y+1);
                  vScrollBar->setValue(Y);
                  QuickLen=0;
                  break;

             case kbPgDn:
                  scrollTo(X,Y+size.y-1);
                  Y=cursor.y+delta.y;
                  if (selectMode)
                     UpdateSelection(cursor.x+delta.x+1,Y+1);
                  vScrollBar->setValue(Y);
                  QuickLen=0;
                  break;

             default:
                 int Key=toupper(event.keyDown.charScan.charCode);
                 if ((Key>='A' && Key<='Z') || Key==' ')
                   {
                    if (QuickLen<MAX_NODE_NAME)
                      {
                       QuickSearch[QuickLen++]=Key;
                       if (topic->SearchNextWith(QuickSearch,QuickLen,selected,QuickVisPos))
                          makeSelectVisible();
                       else
                          QuickLen--;
                      }
                   }
                 else if (Key>='1' && Key<='9')
                   {
                    Key-='1';
                    if (Key<topic->numRefs)
                      {
                       switchToTopic(topic->crossRefs[Key].Name);
                       QuickLen=0;
                      }
                    else
                       return;
                   }
                 else
                   {
                    QuickLen=0;
                    switch (Key)
                      {
                       case kbCtrlL:
                            makeSearch();
                            break;

                       case kbCtrlS:
                            find();
                            break;

                       case kbCtrlN:
                            switchToTopic(topic->Next);
                            break;

                       case kbCtrlP:
                            switchToTopic(topic->Prev);
                            break;

                       case kbCtrlU:
                            switchToTopic(topic->Up);
                            break;

                       case kbCtrlD:
                            switchToTopic("(dir)");
                            break;

                       case kbCtrlT:
                            switchToTopic("Top");
                            break;

                       case '0':
                            if (topic->numRefs)
                               switchToTopic(topic->crossRefs[topic->numRefs-1].Name);
                            break;

                       case kbCtrlG:
                            {
                             TFindInfDialogRec findRec( findStr, findType, findWhere );

                             if ( execDialog( createGoToDialog(), &findRec ) != cmCancel )
                                switchToTopic(findRec.find);
                            }
                            break;
                            
                       case kbCtrlH:
                            switchToTopic("(infview)");
                            break;

                       case kbCtrlO:
                            {
                             int cant=hFile->index->size,i;

                             if (cant)
                               {
                                NameOfNode *ind=hFile->index->Names;
                                TNoCaseStringCollection *NodeList = new TNoCaseStringCollection(20,5);

                                for (i=0; i<cant; i++)
                                    NodeList->insert(ind[i]);
                               
                                struct TListBoxRec
                                {
                                    TCollection *items;
                                    ccIndex selection;
                                } br;
                               
                                br.items=NodeList;
                                br.selection=0;
                               
                                if ( execDialog( createNodeChooseDialog(), &br )!=cmCancel )
                                   switchToTopic((char *)NodeList->at(br.selection));

                                delete NodeList;
                               }
                             break;
                            }

                       default:
                            return;
                      }
                   }
             }
         drawView();
         clearEvent(event);
         break;

     case evMouseDown:
#ifndef NO_DOUBLE_CLICK
          if ( event.mouse.doubleClick )
     	     selectMode |= smDouble;
#endif
          do
            {
             lock();
             MoveToMouse(event.mouse.where,selectMode);
             selectMode |= smExtend;
             unlock();
            }
          while( mouseEvent(event, evMouseMove + evMouseAuto) );
          break;

     /*case evBroadcast:
          switch( event.message.command )
            {
             case cmScrollBarChanged:
                  //checkScrollBar( event, hScrollBar, delta.x );
                  //checkScrollBar( event, vScrollBar, delta.y );
                  break;

             default:
                  return;
            }*/
     }
}


int StrNCmp(const char *s1, const char *s2, size_t n)
{

  if (n == 0) return 0;
  do
   {
    if (toupper(*s1) != *s2++) return 1;
    if (*s1++ == 0)
      break;
   }
  while (--n != 0);
  return 0;
}


char *ScanStr(char *s, char *find, int &Linea, int &Col)
{
  char c, sc;
  size_t len;
  char Buf[256];

  strcpy(Buf,find);
  strupr(Buf);
  find=Buf;

  if ((c = *find++) != 0)
  {
    len = strlen(find);
    do
      {
      do
        {
	 if ((sc = *s++) == 0) return 0;
         if (sc=='\n')
           {
            Col=0;
            Linea++;
           }
         else
            Col++;
        }
      while (toupper(sc) != c);
      }
    while (StrNCmp(s, find, len) != 0);
    s--;
  }
  Col--;
  return s;
}

char *ScanStrCase(char *s, char *find, int &Linea, int &Col)
{
  char c, sc;
  size_t len;

  if ((c = *find++) != 0)
  {
    len = strlen(find);
    do
      {
      do
        {
	 if ((sc = *s++) == 0) return 0;
         if (sc=='\n')
           {
            Col=0;
            Linea++;
           }
         else
            Col++;
        }
      while (sc != c);
      }
    while (strncmp(s, find, len) != 0);
    s--;
  }
  Col--;
  return s;
}



TDialog *createFindDialog();
TDialog *createInfFindDialog();

int TInfViewer::searchInCurrentTopic(int Linea,int Col,int largo)
{
 TPoint Punto;

 do
  {
   lastSearch=findType & fitCaseSens ?
              ScanStrCase(lastSearch,findStr,Linea,Col) :
              ScanStr(lastSearch,findStr,Linea,Col);
   if (lastSearch!=0) // hit and
     {
      if ( !(findType & fitWWord) ||  // (!Whole word or
         // is a whole word)
         (!isalpha(*(lastSearch-1)) && !isalpha(*(lastSearch+largo))) )
        {
         LineEndSelect=LineStartSelect=Linea+1;
         ColStartSelect=Col;
         ColEndSelect=Col+largo;
         Punto.x=Col;
         Punto.y=Linea+1;
         makeVisible(Punto,largo);
         TextSelected=True;
         lastSearch+=largo;
         return 1;
        }
      lastSearch+=largo;
      Col+=largo;
     }
  }
 while (lastSearch);

 return 0;
}


int TInfViewer::searchInCurrentFile(int Linea,int Col,int largo)
{
 char Buf[256];
 char findUp[256];
 FILE *f;

 // first in the current topic
 if (TopicInSearch!=fitNotInTopic && searchInCurrentTopic(Linea,Col,largo))
    return 1;
 // now across topics
 long PosTopic;
 strcpy(findUp,findStr);
 strupr(findUp);
 char *s;
 int LastTopic=TopicInSearch;
 while ((PosTopic=hFile->index->position(++TopicInSearch))!=-1)
   {
    hFile->fSeek(PosTopic);
    topic->ReadNodeInfo(*hFile);
    if (topic->Status)
      {
       switchToTopic("Top");
       break;
      }
    f=hFile->stream;
    do
     {
      if (fgets(Buf,256,f)==NULL) break;
      strupr(Buf);
      s=strstr(Buf,findUp);
     }
    while (s==NULL && Buf[0]!=31);

    if (Buf[0]!=31 && !feof(f))
      {
       switchToTopic(hFile->index->Names[TopicInSearch]);

       lastSearch=topic->Text;
       if (searchInCurrentTopic(0,0,largo))
          return 1;
      }
   };
 if (LastTopic!=fitNotInTopic)
    switchToTopic(hFile->index->Names[LastTopic]);
 return 0;
}

void TInfViewer::makeSearch()
{
 int Linea,Col;
 int largo,Eureka=0;

 if (!SearchArmed) return;

 largo=strlen(findStr);
 if (largo==0) return;

 // Si finaliz una bsqueda antes y ahora presion Ctrl+L que vuelva al ppio.
 if (lastSearch==NULL)
    lastSearch=topic->Text;

 if (lastSearch!=topic->Text)
   {
    Linea=LineEndSelect-1;
    Col=ColEndSelect;
   }
 else
   {
    Linea=0;
    Col=0;
   }

 OpenVerbose=0;
 switch (findWhere)
  {
   case fitInTopic:
        // Search in the current topic, is only in a large buffer
        Eureka=searchInCurrentTopic(Linea,Col,largo);
        break;

   case fitInFile:
        // It's complex; and calls searchInCurrentTopic
        Eureka=searchInCurrentFile(Linea,Col,largo);
        break;

   case fitInDir:
        // It's Waco, and infected (i kill thousands of bugs here).
        if (FileInSearch==fitNotInTopic)
          { // It's the first time
           // Keep track were start
           AddFileToNode(OriginalPlace.Name,topic->Node,hFile->NameOfFile);
           OriginalPlace.HisPos = cursor + delta;
           OriginalPlace.HisSel=selected;
          }
        else
          { // Not the first time, continue in this file
           if (searchInCurrentFile(Linea,Col,largo))
             {
              OpenVerbose=1;
              return;
             }
          }
        // now across files
        // Go to Top of Top
        switchToTopic("(dir)Top");
        while ((++FileInSearch)<topic->numRefs)
          {
           switchToTopic(topic->crossRefs[FileInSearch].Name);
           if (!hFile->Status)
             {
              TopicInSearch=fitNotInTopic;
              if (searchInCurrentFile(Linea,Col,largo))
                {
                 OpenVerbose=1;
                 return;
                }
             }
           switchToTopic("(dir)Top");
          }
        switchToTopic(&OriginalPlace);
        break;
  }
 if (!Eureka)
   {
    messageBox(_("Sorry can't find the string"),mfError | mfOKButton);
    SearchArmed=0;
   }
 OpenVerbose=1;
}


void TInfViewer::find()
{
 TFindInfDialogRec findRec( findStr, findType, findWhere );
 lastSearch=topic->Text;
 FileInSearch=TopicInSearch=fitNotInTopic;
 if ( execDialog( createInfFindDialog(), &findRec ) != cmCancel )
   {
    strcpy( findStr, findRec.find );
    findType = findRec.options;
    findWhere =  findRec.ops2;
    SearchArmed=1;
    makeSearch();
   }
}


/***************************************************************

 class TInfWindow

 Just like THelpWindow.

***************************************************************/

TInfWindow::TInfWindow( TInfFile *hFile, char *context, char *match,
                        void (*ir)(char *b, long l),Boolean IsTheOne):
       TWindow( TRect(0,0,50,18), "InfView", wnNoNumber ),
       TWindowInit( &TInfWindow::initFrame)
{
 TRect r(0, 0, 50, 18);
 options = (options | ofCentered);
 r.grow(-1,-1);

 viewer=new TInfViewer (r,standardScrollBar(sbHorizontal | sbHandleKeyboard),
   standardScrollBar(sbVertical | sbHandleKeyboard), hFile, context,
   match);
 viewer->InsertRoutine=ir;
 insert(viewer);
 helpCtx = hcInfView;
 isTheOne=IsTheOne;
}

TPalette& TInfWindow::getPalette() const
{
    static TPalette palette(cInfWindow, sizeof( cInfWindow)-1);
    return palette;
}

void TInfWindow::close()
{
 if (isTheOne)
    hide();
 else
    TWindow::close();
}


/***************************************************************

 class TStrListBox

 Una ListBox que responde a Enter y Doble click.
 A List Box with Enter+Double click response.

 La respuesta es un mensaje cmStrChoose y en infoPtr el ndice
seleccionado.

***************************************************************/

void TStrListBox::handleEvent(TEvent& event)
{
 short a=focused;
#ifdef NO_DOUBLE_CLICK
 if ( event.what==evKeyDown && event.keyDown.keyCode==kbEnter )
#else
 if ( (event.what==evKeyDown && event.keyDown.keyCode==kbEnter)
    || (event.what == evMouseDown && event.mouse.doubleClick) )
#endif
   {
    message( owner, evBroadcast, cmStrChoose, (void *)&focused );
    clearEvent(event);
    return;
   }
 TListBox::handleEvent(event);
 if (focused!=a)
    draw();
}

/***************************************************************

 class THelpControlDialog

 Una caja de dilogo para controlar el help.
 A Dialog to control the help.

***************************************************************/

void THelpControlDialog::handleEvent(TEvent& event)
{
 TDialog::handleEvent(event);
 if( event.what == evCommand || event.what == evBroadcast)
  {
   //ushort comm=event.message.command;
   switch( event.message.command )
           {
            // a button
            case chcdNext:
            case chcdPrev:
            case chcdUp:
            case chcdPrevH:
            case chcdHide:
                {
                endModal(event.message.command);
                clearEvent(event);
                }
                break;

            // from the list
            case cmStrChoose:
                {
                 endModal(chcdHistSel+*((ushort *)event.message.infoPtr));
                 clearEvent(event);
                }
                break;
            default:
                break;
           }
  }
}


THelpControlDialog::THelpControlDialog(TCollectionNoSort *oCol) :
    TDialog( TRect( 1, 1, 51, 15 ), _("Help Control") ),
    TWindowInit( &THelpControlDialog::initFrame )
{
 TRect r(1,2,18,4);
 insert( new TButton(r,_("~N~ext"),100,bfDefault) );
 r.a.y+=2;
 r.b.y+=2;
 insert( new TButton(r,_("~P~revious"),101,bfNormal) );
 r.a.y+=2;
 r.b.y+=2;
 insert( new TButton(r,_("~U~p"),102,bfNormal) );
 r.a.y+=2;
 r.b.y+=2;
 insert( new TButton(r,_("Prev. ~H~ist."),103,bfNormal) );
 r.a.y+=2;
 r.b.y+=2;
 insert( new TButton(r,_("H~i~de"),104,bfNormal) );

 TScrollBar *sb = new TScrollBar( TRect(47,3,48,12) );
 insert(sb);
 TStrListBox *ListaH = new TStrListBox( TRect(21,3,47,12),(ushort)1, sb );
 insert(ListaH);
 ListaH->newList(oCol);
 insert (new TLabel(TRect(21,2,47,3),_("Hi~s~tory"), ListaH));

 options|=ofTopSelect;
 selectNext( False );
 helpCtx=hcInfControl;
}


TDialog *createInfFindDialog()
{
 TDialog *d = new TDialog( TRect( 0, 0, 38, 15 ), _("Inf Find") );

 d->options |= ofCentered;

 TInputLine *control = new TInputLine( TRect( 3, 3, 32, 4 ), 80 );
 d->insert( control );
 d->insert(
     new TLabel( TRect( 2, 2, 30, 3 ), _("~T~ext to find"), control ) );
 d->insert(
     new THistory( TRect( 32, 3, 35, 4 ), control, 10 ) );

 d->insert( new TCheckBoxes( TRect( 3, 5, 35, 7 ),
     new TSItem( _("~C~ase sensitive"),
     new TSItem( _("~W~hole words only"), 0 ))));

 d->insert( new TRadioButtons( TRect( 3, 8, 35, 11 ),
     new TSItem( _("In current ~t~opic"),
     new TSItem( _("In the ~f~ile"),
     new TSItem( _("In ~a~ll files"), 0 )))));

 d->insert(
     new TButton( TRect( 14, 12, 24, 14 ), _("~O~K"), cmOK, bfDefault ) );
 d->insert(
     new TButton( TRect( 26, 12, 36, 14 ), _("Cancel"), cmCancel, bfNormal ) );

 d->selectNext( False );
 d->helpCtx=hcInfSearch;
 return d;
}

TDialog *createGoToDialog()
{
 TDialog *d = new TDialog( TRect( 0, 0, 38, 9 ), _("Go To") );

 d->options |= ofCentered;

 TInputLine *control = new TInputLine( TRect( 3, 3, 32, 4 ), 80 );
 d->insert( control );
 d->insert(
     new TLabel( TRect( 2, 2, 30, 3 ), _("~N~ame of Node"), control ) );
 d->insert(
     new THistory( TRect( 32, 3, 35, 4 ), control, 10 ) );

 d->insert(
     new TButton( TRect( 14, 6, 24, 8 ), _("~O~K"), cmOK, bfDefault ) );
 d->insert(
     new TButton( TRect( 26, 6, 36, 8 ), _("Cancel"), cmCancel, bfNormal ) );

 d->selectNext( False );
 d->helpCtx=hcInfGoto;
 return d;
}

TDialog *createNodeChooseDialog()
{
 TRect r( 0, 0, 60, 20 );
 TDialog *d = new TDialog( r, _("Tag Table") );

 d->options |= ofCentered;

 TScrollBar *sb = new TScrollBar( TRect(r.b.x-3,2,r.b.x-2,r.b.y-4) );
 d->insert(sb);
 TSortedListBox *ListaN = new TSortedListBox( TRect(2,2,r.b.x-3,r.b.y-4),(ushort)1, sb );
 d->insert(ListaN);
 d->insert (new TLabel(TRect(2,1,r.b.x-3,2),_("~N~odes"), ListaN));

 d->insert(
     new TButton( TRect( 2, r.b.y-3, 16, r.b.y-1 ), _("~O~K"), cmOK, bfDefault ) );
 d->insert(
     new TButton( TRect( 20, r.b.y-3, 34, r.b.y-1 ), _("Cancel"), cmCancel, bfNormal ) );
 d->selectNext( False );
 d->helpCtx=hcInfChoose;

 return d;
}

#pragma warn .dsz

