/*

   Keyboard and general definitions.

     This code is based on the Allegro lib, see keyboard.c to get more
   information

 */


#ifndef KYBOARD_H
#define KYBOARD_H

/* That's because I'm forcing to BCC to compile all like C++ regardless the
   extention */
#ifdef __DJGPP__
#ifdef __cplusplus
extern "C" {
#endif
#endif

/*******************************************/
/************ Some global stuff ************/
/*******************************************/

#ifndef TRUE
#define TRUE         -1
#define FALSE        0
#endif

#define END_OF_FUNCTION(x)    void x##_end() { }

#ifdef __cplusplus
#define VarArgList ...
#else
#define VarArgList
#endif


#ifdef __DJGPP__

#include <dpmi.h>
#define LOCK_VARIABLE(x)      _go32_dpmi_lock_data((void *)&x, sizeof(x))
#define LOCK_FUNCTION(x)      _go32_dpmi_lock_code(x, (long)x##_end - (long)x)
#define Inline inline

#else

#define LOCK_VARIABLE(x)
#define LOCK_FUNCTION(x)
#define Inline

#endif

void allegro_exit(void);

/*******************************************/
/************ Keyboard routines ************/
/*******************************************/

int install_keyboard(void);
void remove_keyboard(void);

extern volatile char key[128];

extern int three_finger_flag;

int keypressed(void);
int readkey(void);
void simulate_keypress(int key);
void clear_keybuf(void);

extern char key_ascii_table[];
extern char key_shift_table[];
extern char key_control_table[];
extern volatile int _shifts_flags;

#define LedScrollLock 1
#define LedNumLock    2
#define LedCapsLock   4

#define kbShiftFlg    1
#define kbCtrlFlg     2
#define kbAltFlg      4
#define kbScroLockFlg 8
#define kbNumLockFlg  16
#define kbCapsLockFlg 32
#define kbInAltSeqFlg 64
#define kbSpecialMask 7

#define SCANCODE_TO_KEY(c)       (((c)<<8) + key_ascii_table[c])
#define SCANCODE_TO_CAPS(c)      (((c)<<8) + key_capslock_table[c])
#define SCANCODE_TO_SHIFT(c)     (((c)<<8) + key_shift_table[c])
#define SCANCODE_TO_CONTROL(c)   (((c)<<8) + key_control_table[c])
#define SCANCODE_TO_ALT(c)       ((c)<<8)

#define KEY_ESC         1        /* keyboard scan codes */
#define KEY_1           2
#define KEY_2           3
#define KEY_3           4
#define KEY_4           5
#define KEY_5           6
#define KEY_6           7
#define KEY_7           8
#define KEY_8           9
#define KEY_9           10
#define KEY_0           11
#define KEY_MINUS       12
#define KEY_EQUALS      13
#define KEY_BACKSPACE   14
#define KEY_TAB         15
#define KEY_Q           16
#define KEY_W           17
#define KEY_E           18
#define KEY_R           19
#define KEY_T           20
#define KEY_Y           21
#define KEY_U           22
#define KEY_I           23
#define KEY_O           24
#define KEY_P           25
#define KEY_OPENBRACE   26
#define KEY_CLOSEBRACE  27
#define KEY_ENTER       28
#define KEY_CONTROL     29
#define KEY_A           30
#define KEY_S           31
#define KEY_D           32
#define KEY_F           33
#define KEY_G           34
#define KEY_H           35
#define KEY_J           36
#define KEY_K           37
#define KEY_L           38
#define KEY_COLON       39
#define KEY_QUOTE       40
#define KEY_TILDE       41
#define KEY_LSHIFT      42
#define KEY_Z           44
#define KEY_X           45
#define KEY_C           46
#define KEY_V           47
#define KEY_B           48
#define KEY_N           49
#define KEY_M           50
#define KEY_COMMA       51
#define KEY_STOP        52
#define KEY_SLASH       53
#define KEY_RSHIFT      54
#define KEY_ASTERISK    55
#define KEY_ALT         56
#define KEY_SPACE       57
#define KEY_CAPSLOCK    58
#define KEY_F1          59
#define KEY_F2          60
#define KEY_F3          61
#define KEY_F4          62
#define KEY_F5          63
#define KEY_F6          64
#define KEY_F7          65
#define KEY_F8          66
#define KEY_F9          67
#define KEY_F10         68
#define KEY_NUMLOCK     69
#define KEY_SCRLOCK     70
#define KEY_HOME        71
#define KEY_UP          72
#define KEY_PGUP        73
#define KEY_MINUS_PAD   74
#define KEY_LEFT        75
#define KEY_5_PAD       76
#define KEY_RIGHT       77
#define KEY_PLUS_PAD    78
#define KEY_END         79
#define KEY_DOWN        80
#define KEY_PGDN        81
#define KEY_INSERT      82
#define KEY_DEL         83
#define KEY_F11         87
#define KEY_F12         88
#define KEY_PAD        100
#define MACRO          111

#ifdef __DJGPP__
#ifdef __cplusplus
}
#endif
#endif

#endif          /* ifndef KYBOARD_H */

